/**
 * @module package/quiqqer/authmetamask/bin/controls/Login
 * @author www.pcsg.de (Henning Leutz)
 */
define('package/quiqqer/authmetamask/bin/controls/Login', [

    'qui/QUI',
    'qui/controls/Control',
    'qui/controls/loader/Loader',
    'qui/controls/windows/Confirm',

    'Ajax',
    'Locale',

    'css!package/quiqqer/authmetamask/bin/controls/Login.css',

], function(QUI, QUIControl, QUILoader, QUIConfirm, QUIAjax, QUILocale) {
    'use strict';

    const lg = 'quiqqer/authmetamask';

    return new Class({

        Extends: QUIControl,
        Type: 'package/quiqqer/authmetamask/bin/controls/Login',

        Binds: [
            '$onImport',
            'click',
        ],

        options: {
            uid: false,  // QUIQQER User ID
        },

        initialize: function(options) {
            this.parent(options);

            this.addEvents({
                onImport: this.$onImport,
            });

            this.Loader = new QUILoader();
            this.$nonce = null;
        },

        /**
         * @return {HTMLDivElement}
         */
        create: function() {
            this.$Elm = new Element('div', {
                'class': 'quiqqer-auth-metamask-login',
                'html': '<div class="quiqqer-auth-metamask-login-info"></div>' +
                    '<img class="quiqqer-auth-metamask-login-icon" ' +
                    '   src="' + window.URL_OPT_DIR +
                    'quiqqer/authmetamask/bin/MetaMaskLogo.png" ' +
                    '/>' +
                    '<div class="quiqqer-auth-metamask-login-btns">' +
                    '   <span>Register now with MetaMask</span>' +
                    '</div>',
            });

            this.$Elm.addEvent('click', this.click);
            this.Loader.inject(this.$Elm);

            return this.$Elm;
        },

        /**
         * Event: onImport
         */
        $onImport: function() {
            this.$Input = this.getElm();
            this.$Input.type = 'hidden';
            this.$Form = this.$Input.getParent('form');
            this.$autoLogin = this.$Input.get('data-autologin') === '1';

            this.$Form.addClass('quiqqer-metamask-login-button');

            this.$Form.getElement('.quiqqer-metamask-login').set({
                'data-quiid': this.getId(),
                'data-qui': this.getType(),
            });

            this.create().inject(this.$Input, 'after');
        },

        /**
         * login click
         */
        click: function() {
            let web3;
            let self = this;

            this.Loader.show();

            this.getWeb3().then(function(w3) {
                web3 = w3;
                return self.getLogin();
            }).then(function() {
                return self.signing();
            }).then(function(signature) {

                const address = web3.currentProvider.selectedAddress;
                const host = window.location.host;

                return self.auth(address, signature, host);
            }).then(function(authenticated) {
                if (authenticated) {
                    self.Loader.hide();
                    window.location.reload();
                    return;
                }

                self.Loader.hide();
            }).catch(function(err) {
                console.error(err);

                if (typeof err.getMessage === 'function') {
                    QUI.getMessageHandler().then(function(MH) {
                        MH.addError(err.getMessage());
                    });
                }

                self.Loader.hide();
            });
        },

        /**
         * get web3 client
         *
         * @return {Promise}
         */
        getWeb3: function() {
            return new Promise(function(resolve) {
                require(['package/quiqqer/authmetamask/bin/Web3'],
                    function(Web3) {
                        Web3.getWeb3().then(resolve);
                    });
            });
        },

        /**
         * @return {Promise}
         */
        getLogin: function() {
            const self = this;

            return this.getWeb3().then(function(web3) {
                return web3.currentProvider.selectedAddress;
            }).then(function(address) {
                return new Promise(function(resolve) {
                    QUIAjax.get('package_quiqqer_authmetamask_ajax_login',
                        function(nonce) {
                            self.$nonce = nonce;
                            resolve(nonce);
                        }, {
                            'package': 'quiqqer/authmetamask',
                            address: address,
                        });
                });
            });
        },

        /**
         * Signing message
         *
         * @return {Promise}
         */
        signing: function() {
            const self = this;

            return this.getWeb3().then(function(web3) {
                const address = web3.currentProvider.selectedAddress;

                return new Promise(function(resolve, reject) {
                    let message = QUILocale.get(lg, 'signing.message', {
                        nonce: self.$nonce,
                        host: window.location.host,
                    });

                    message = web3.utils.fromUtf8(message);

                    web3.eth.personal.sign(
                        message,
                        address,
                        '',
                        function(err, signature) {
                            if (err) {
                                return reject(err);
                            }

                            return resolve(signature);
                        },
                    );
                });
            });
        },

        /**
         *
         * @param address
         * @param signature
         * @param host
         * @return {*}
         */
        auth: function(address, signature, host) {
            return new Promise(function(resolve, reject) {
                QUIAjax.post('package_quiqqer_authmetamask_ajax_auth', resolve,
                    {
                        'package': 'quiqqer/authmetamask',
                        onError: reject,
                        address: address,
                        signature: signature,
                        host: host,
                    });
            });
        },
    });
});

