<?php

/**
 * This file contains QUI\Auth\Google\Auth
 */

namespace QUI\Auth\MetaMask;

use Elliptic\EC;
use kornrunner\Keccak;
use QUI;
use QUI\Control;
use QUI\Database\Exception;
use QUI\FrontendUsers\Exception\UserAlreadyExistsException;
use QUI\Locale;
use QUI\Users\AbstractAuthenticator;
use QUI\Users\User;

use function hex2bin;
use function is_null;
use function ord;
use function strlen;
use function strtolower;
use function substr;
use function uniqid;

/**
 * Class Auth
 *
 * Authentication handler for Google authentication
 *
 * @package QUI\Authe\Google2Fa
 */
class Auth extends AbstractAuthenticator
{
    /**
     * User that is to be authenticated
     *
     * @var QUI\Interfaces\Users\User | null
     */
    protected QUI\Interfaces\Users\User | null $User = null;

    /**
     * Auth Constructor.
     *
     * @param array|integer|string $user - name of the user, or user id
     */
    public function __construct(array | int | string $user = '')
    {
        if (!empty($user)) {
            try {
                $this->User = QUI::getUsers()->getUserByName($user);
            } catch (\Exception) {
                $this->User = QUI::getUsers()->getNobody();
            }
        }
    }

    /**
     * @param null|Locale $Locale
     * @return string
     */
    public function getTitle(null | Locale $Locale = null): string
    {
        if (is_null($Locale)) {
            $Locale = QUI::getLocale();
        }

        return $Locale->get('quiqqer/authmetamask', 'authmetamask.title');
    }

    /**
     * @param null|Locale $Locale
     * @return string
     */
    public function getDescription(null | Locale $Locale = null): string
    {
        if (is_null($Locale)) {
            $Locale = QUI::getLocale();
        }

        return $Locale->get('quiqqer/authmetamask', 'authmetamask.description');
    }

    /**
     * Authenticate the user
     *
     * @param string|array|integer $authData
     *
     * @throws Exception
     * @throws Exception
     * @throws QUI\Exception
     * @throws QUI\FrontendUsers\Exception
     * @throws UserAlreadyExistsException
     */
    public function auth(string | array | int $authData): void
    {
        $message = $authData['message'];
        $signature = $authData['signature'];
        $address = $authData['address'];

        try {
            $verify = QUI\Auth\MetaMask\Auth::verifySignature($message, $signature, $address);
        } catch (\Exception) {
            throw new QUI\Auth\MetaMask\Exception('Authentication failed');
        }

        if (!$verify) {
            throw new QUI\Auth\MetaMask\Exception('Authentication failed');
        }

        try {
            $userId = QUI\MetaMask\MetaMask::getUserIdByEthAddress($address);

            if (!$userId) {
                throw new QUI\Exception('User not found', 404);
            }
        } catch (\Exception) {
            if (QUI::getUserBySession()->getId()) {
                // if user is already logged in, we can add the wallet to the user
                QUI::getDataBase()->update(
                    QUI\MetaMask\MetaMask::table(),
                    ['userId' => QUI::getUserBySession()->getId()],
                    ['address' => $address]
                );
            } else {
                // user not found
                // if no user, create it
                $_POST['registration'] = true;
                $_POST['termsOfUseAccepted'] = true;

                $MetaMaskRegistrar = new QUI\Registration\MetaMask\Registrar();
                $Registration = new QUI\FrontendUsers\Controls\Registration();
                $Registration->setAttribute('Registrar', $MetaMaskRegistrar);
                $Registration->register();

                $User = $Registration->getRegisteredUser();

                if (!$User) {
                    throw new QUI\Exception('User not found', 404);
                }


                try {
                    if (class_exists('QUI\ERP\Customer\Customers')) {
                        $User->addToGroup(QUI\ERP\Customer\Customers::getInstance()->getCustomerGroupId());
                    }
                } catch (QUI\Exception) {
                }

                try {
                    if ($User->getName() !== $address) {
                        $User->setAttribute('username', $address);
                        $User->save(QUI::getUsers()->getSystemUser());
                    }
                } catch (QUI\Exception) {
                }

                try {
                    $User->activate('', QUI::getUsers()->getSystemUser());
                } catch (QUI\Exception) {
                }

                QUI::getDataBase()->update(
                    QUI\MetaMask\MetaMask::table(),
                    ['userId' => $User->getId()],
                    ['address' => $address]
                );
            }
        }

        $this->setSession($address);
        QUI::getSession()->set('auth-globals', 1);
    }

    /**
     * @param $address
     *
     * @throws Exception
     * @throws QUI\Exception
     * @throws QUI\FrontendUsers\Exception
     * @throws QUI\Users\Exception
     * @throws UserAlreadyExistsException
     * @throws Exception
     */
    public function setSession($address): void
    {
        try {
            $userId = QUI\MetaMask\MetaMask::getUserIdByEthAddress($address);
        } catch (\Exception) {
            throw new QUI\Auth\MetaMask\Exception('Authentication failed');
        }

        $User = null;

        if (!empty($userId)) {
            try {
                $User = QUI::getUsers()->get($userId);
            } catch (QUI\Exception) {
            }
        }

        // user has no started registration
        if ($User === null) {
            $FrontendUsers = new QUI\FrontendUsers\Handler();
            $Registrars = $FrontendUsers->getRegistrars();

            $Registrars = $Registrars->filter(function ($R) {
                return $R instanceof QUI\Registration\MetaMask\Registrar;
            });

            /* @var $Registrar QUI\Registration\MetaMask\Registrar */
            $Registrar = $Registrars->get(0);

            $_POST['registrar'] = $Registrar->getHash();
            $_POST['registration'] = 1;
            $_POST['termsOfUseAccepted'] = 1;

            $Registration = new QUI\FrontendUsers\Controls\Registration();
            $Registration->setAttribute('Registrar', $Registrar);
            $Registration->register();

            // check if name already exists
            $User = $Registration->getRegisteredUser();

            if (!QUI::getUsers()->usernameExists($address)) {
                $User->setAttribute('username', $address);
                $User->save(QUI::getUsers()->getSystemUser());
            }

            QUI::getDataBase()->update(
                QUI\MetaMask\MetaMask::table(),
                [
                    'userId' => $User->getId(),
                    'nonce' => uniqid() // reset nonce, because of security
                ],
                ['address' => $address]
            );
        }

        $Session = QUI::getSession();
        $Session->set('uid', $User->getId());
        $Session->set('username', $User->getName());
        $Session->set('auth-globals', 1);

        QUI::getUsers()->login();
    }

    /**
     * Return the user object
     *
     * @return \QUI\Interfaces\Users\User
     */
    public function getUser(): QUI\Interfaces\Users\User
    {
        return $this->User;
    }

    /**
     * Return the quiqqer user id
     *
     * @return int
     */
    public function getUserId(): int
    {
        return $this->User->getId();
    }

    /**
     * @return Control|null
     */
    public static function getLoginControl(): ?QUI\Control
    {
        return new QUI\Auth\MetaMask\Controls\Login();
    }

    /**
     * @return Control|null
     */
    public static function getPasswordResetControl(): ?QUI\Control
    {
        return null;
    }

    /**
     * @param $message
     * @param $signature
     * @param $address
     *
     * @return bool
     *
     * @throws \Exception
     */
    public static function verifySignature($message, $signature, $address): bool
    {
        $msgLen = strlen($message);

        $hash = Keccak::hash(
            "\x19Ethereum Signed Message:\n{$msgLen}{$message}",
            256
        );

        $sign = [
            "r" => substr($signature, 2, 64),
            "s" => substr($signature, 66, 64)
        ];

        $recId = ord(hex2bin(substr($signature, 130, 2))) - 27;

        if ($recId != ($recId & 1)) {
            return false;
        }

        $EC = new EC('secp256k1');
        $pubKey = $EC->recoverPubKey($hash, $sign, $recId);

        return strtolower($address) == strtolower(self::pubKeyToAddress($pubKey));
    }

    /**
     * @param $pubKey
     * @return string
     *
     * @throws \Exception
     */
    public static function pubKeyToAddress($pubKey)
    {
        return "0x" . substr(Keccak::hash(substr(hex2bin($pubKey->encode("hex")), 1), 256), 24);
    }
}
